#!/usr/bin/env node
"use strict";
/**
 * typescript-build-linker
 * Link together TypeScript packages in a monorepo
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const path = __importStar(require("path"));
const t = __importStar(require("io-ts"));
const A = __importStar(require("fp-ts/Array"));
const E = __importStar(require("fp-ts/Either"));
const O = __importStar(require("fp-ts/Option"));
const T = __importStar(require("fp-ts/Task"));
const R = __importStar(require("fp-ts/Record"));
const S = __importStar(require("fp-ts/Set"));
const IO = __importStar(require("fp-ts/IO"));
const TE = __importStar(require("fp-ts/TaskEither"));
const Console = __importStar(require("fp-ts/Console"));
const D = __importStar(require("io-ts-docopt"));
const io_ts_docopt_1 = require("io-ts-docopt");
const Semigroup_1 = require("fp-ts/lib/Semigroup");
const Apply_1 = require("fp-ts/Apply");
const PathReporter_1 = require("io-ts/lib/PathReporter");
const function_1 = require("fp-ts/function");
const lerna_utils_1 = require("@typescript-tools/lerna-utils");
const stringify_json_1 = require("@typescript-tools/stringify-json");
const lerna_packages_1 = require("@typescript-tools/lerna-packages");
const monorepo_root_1 = require("@typescript-tools/monorepo-root");
const TsConfig_1 = require("@typescript-tools/io-ts/dist/lib/TsConfig");
const dependency_graph_1 = require("@typescript-tools/dependency-graph");
const withFallback_1 = require("io-ts-types/lib/withFallback");
const StringifiedJSON_1 = require("@typescript-tools/io-ts/dist/lib/StringifiedJSON");
const trace_1 = require("@strong-roots-capital/trace");
const get_relative_path_1 = __importDefault(require("get-relative-path"));
const debug_1 = __importDefault(require("debug"));
const fast_deep_equal_1 = __importDefault(require("fast-deep-equal"));
const ts_pattern_1 = require("ts-pattern");
const Eq_1 = require("fp-ts/lib/Eq");
const debug = {
    cmd: debug_1.default('link')
};
const docstring = `
Usage:
    typescript-build-linker [<repository>]

Options:
    <repository>    Path to monorepo root, defaults to cwd
`;
const CommandLineOptions = io_ts_docopt_1.withEncode(t.type({
    '<repository>': withFallback_1.withFallback(t.string, process.cwd()),
}), a => ({
    repository: a['<repository>'],
}));
// Widens the type of a particular Err into Err
const err = function_1.identity;
const monorepoRoot = function_1.flow(monorepo_root_1.monorepoRoot, E.mapLeft(err), TE.fromEither);
const lernaPackages = function_1.flow(lerna_packages_1.lernaPackages, TE.mapLeft(err));
const dependencyGraph = (root) => function_1.pipe(dependency_graph_1.dependencyGraph(root, { recursive: false }), TE.mapLeft(err));
const decodeDocopt = function_1.flow(D.decodeDocopt, E.mapLeft(errors => PathReporter_1.PathReporter.report(E.left(errors)).join('\n')), E.mapLeft(error => err({ type: 'docopt decode', error })), TE.fromEither);
const readFile = (filename) => function_1.pipe(lerna_utils_1.readFile(filename), TE.mapLeft(error => err({ type: 'unable to read file', filename, error })));
const decodeFile = (codec) => (filename) => (contents) => function_1.pipe(StringifiedJSON_1.StringifiedJSON(codec).decode(contents), E.mapLeft(errors => PathReporter_1.PathReporter.report(E.left(errors)).join('\n')), E.mapLeft(error => err({ type: 'unexpected file contents', filename, error })), TE.fromEither);
const stringifyJSON = (onError) => (json) => function_1.pipe(stringify_json_1.stringifyJSON(onError)(json), E.mapLeft(error => err({ type: 'unable to stringify json', json, error })), TE.fromEither);
const writeFile = (filename) => (contents) => function_1.pipe(contents, trace_1.trace(debug.cmd, `Writing file ${filename}`), lerna_utils_1.writeFile(filename), TE.mapLeft(error => err({ type: 'unable to write file', filename, error })));
const writeJson = (filename) => (value) => function_1.pipe(stringifyJSON(E.toError)(value), TE.chain(writeFile(filename)));
const writeJsonIfModified = (filename, original, desired) => function_1.pipe(O.fromPredicate((value) => !fast_deep_equal_1.default(original, value))(desired), O.fold(function_1.constant(TE.right(undefined)), function_1.constant(writeJson(filename)(desired))));
const mapToRecord = (map) => Array
    .from(map.entries())
    .reduce((acc, [key, value]) => Object.assign(acc, { [key]: value }), {});
const parentDirectory = (directory) => ts_pattern_1.match(directory)
    .with('.', function_1.constant(O.none))
    .otherwise(() => O.some(path.dirname(directory)));
const ancestors = (directory) => {
    const ancestors = [];
    let parent = parentDirectory(directory);
    while (O.isSome(parent)) {
        function_1.pipe(parent, O.map(parent_ => {
            ancestors.push([parent_, path.basename(directory)]);
            directory = parent_;
            parent = parentDirectory(parent_);
        }));
    }
    return ancestors;
};
const linkChildrenPackages = (root) => function_1.pipe(lernaPackages(root), 
// map parents to children
TE.map(packages => R.fromFoldableMap(A.getMonoid(), A.array)(packages, pkg => [
    // map parent directories into relative paths from root
    root === path.dirname(pkg.location)
        // treat the case where package is a direct child of monorepo-root,
        // because relativePath returns '../<root>' most irregularly
        ? '.'
        : get_relative_path_1.default(root + '/', path.dirname(pkg.location)),
    [pkg]
])), 
// isolate the lerna package directory
TE.map(R.map(A.map(pkg => path.basename(pkg.location)))), 
// create ancestor references
TE.map(packagesByDirectory => R.fromFoldableMap(S.getUnionMonoid(Eq_1.eqString), A.array)(function_1.pipe(Object.entries(packagesByDirectory), A.chain(([directory, packages]) => [
    ...ancestors(directory),
    ...packages.map(pkg => [directory, pkg]),
])), ([directory, pkg]) => [directory, new Set([pkg])])), 
// map set of children packages to set of children project references
TE.map(R.map(packages => Array.from(packages).map(pkg => ({ path: pkg })))), 
// map to write instructions
TE.map(R.mapWithIndex((parentDirectory, references) => {
    const tsconfigFile = path.resolve(root, parentDirectory, 'tsconfig.json');
    return function_1.pipe(readFile(tsconfigFile), TE.orElse(function_1.constant(TE.right('{}'))), TE.chain(decodeFile(TsConfig_1.TsConfig)(tsconfigFile)), 
    // The `files: []` prevents an accidental invocation
    // of `tsc` without `-b` from trying to build the
    // entire directory as one compilation:
    // https://github.com/RyanCavanaugh/learn-a
    TE.chain(tsconfig => writeJsonIfModified(path.resolve(root, tsconfigFile), tsconfig, Object.assign({ files: [] }, tsconfig, { references }))));
})), 
// write in parallel
TE.chain(Apply_1.sequenceS(TE.taskEither)), 
// return void on success
TE.map(function_1.constVoid));
const linkPackageDependencies = (root) => function_1.pipe(
// create map of package name to location (absolute path)
lernaPackages(root), TE.map(packages => R.fromFoldableMap(Semigroup_1.getLastSemigroup(), A.array)(packages, pkg => [pkg.name, pkg.location])), TE.chain(lernaPackages => function_1.pipe(dependencyGraph(root), TE.map(mapToRecord), TE.map(R.reduceWithIndex({}, (packageName, acc, dependencies) => Object.assign(acc, {
    [lernaPackages[packageName]]: dependencies
        .map(d => d.location)
        .map(p => path.relative(lernaPackages[packageName], p))
        .map(path => ({ path }))
}))))), TE.map(R.mapWithIndex((packageDirectory, references) => {
    const tsconfigFile = path.resolve(packageDirectory, 'tsconfig.json');
    return function_1.pipe(readFile(tsconfigFile), TE.chain(decodeFile(TsConfig_1.TsConfig)(tsconfigFile)), TE.chain(tsconfig => writeJsonIfModified(tsconfigFile, tsconfig, Object.assign({}, tsconfig, { references }))));
})), 
// write in parallel
TE.chain(Apply_1.sequenceS(TE.taskEither)), 
// return void on success
TE.map(function_1.constVoid));
const exit = (code) => () => process.exit(code);
const main = function_1.pipe(decodeDocopt(CommandLineOptions, docstring), TE.chain(options => monorepoRoot(options.repository)), TE.chain(root => function_1.pipe([
    linkChildrenPackages(root),
    linkPackageDependencies(root),
], TE.sequenceArray)), TE.fold(function_1.flow(Console.error, IO.chain(() => exit(1)), T.fromIO), function_1.constant(T.of(undefined))));
main();
// Local Variables:
// mode: typescript
// End:
//# sourceMappingURL=index.js.map